#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2010-2019 Centrify Corporation. All rights reserved.
#
# Machine mapper script to configure DirectAudit settings.
#
#  This script will configure DirectAudit settings by performing the following 
#  actions:
#  1. Copy DirectAudit settings to /var/centrifyda
#  2. Set audit commands
#
#
#  Copy DirectAudit settings to /var/centrifyda
#  --------------------------------------------
#  For DA installation GP.
#
#
#  Set audit commands
#  ------------------
#  This script will configure the commands to be audited according to the GP 
#  settings. A list of commands can be defined on GP to indicate the commands 
#  to be audited and NOT to be audited. Commands to be audited are added to the 
#  list with action "Enable", and commands NOT to be audited are added with 
#  action "Disable".
#
#  Note that we cannot control commands not added to the list. For example, if 
#  admin enables command "/usr/sbin/abc" manually on a UNIX machine without 
#  adding the command on GP UI, this script will do nothing about this command.
#  
#
#  Map:     Configure DirectAudit settings.
#
#  Unmap:   Reset DirectAudit settings.
#
# Parameters: <map|unmap> [username] mode
#   map|unmap   action to take
#   username    username (not used)
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#
##############################################################################

use strict;

use lib "/usr/share/centrifydc/perl";
use CentrifyDC::Include;
use File::Copy;
use CentrifyDC::GP::General qw(:debug IsEmpty RunCommand);
use CentrifyDC::GP::Args;
use CentrifyDC::GP::RegHelper;


# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


my $OLD_DA_GP_POLICY_DIR = "/var/centrifydc/reg/machine/software/policies/centrify/directaudit/agent";
my $NEW_DA_GP_POLICY_DIR = "/var/centrifydc/reg/machine/software/policies/centrify/directaudit/common";
my $FILENAME = "Registry.pol";
my $DACONTROL = "/usr/sbin/dacontrol";
my $DARELOAD = "/usr/sbin/dareload";
my $DAINFO = "/usr/bin/dainfo";

my $da_gp_policy_file = "";
my $is_policy_file_exist = 0;

my $AUDIT_COMMANDS_REGKEY = "software/policies/centrify/centrifyda/settings/unixagent/auditcommands";
my $AUDIT_COMMANDS_VALUE_ENABLE = "Enable";
my $AUDIT_COMMANDS_VALUE_DISABLE = "Disable";

my $SESSION_AUDIT_REGKEY = "software/policies/centrify/centrifyda/settings/unixagent/sessionaudit";
my $SESSION_AUDIT_NAME = "dad.session.audit.enabled";
my $SESSION_AUDIT_VALUE_ENABLED = "true";
my $SESSION_AUDIT_VALUE_DISABLED = "false";

my $AUDIT_ENABLE_REGKEY = "software/policies/centrify/centrifyda/settings/unixagent/advancedaudit";
my $ADVANCED_MONITOR_ENABLED = "dad.event.enabled";
my $ADVANCED_MONITOR_VALUE_ENABLED = "true";
my $ADVANCED_MONITOR_VALUE_DISABLED = "false";
my $ADVANCED_MONITOR_VALUE_NOT_CONF = "";

use constant MAX_RETRIES=>5;

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# Configure DirectAudit session auditing according to GP settings - "Enable DirectAudit session auditing".
#
#   $_[0]:  action  (map/unmap)
#
#   return: 1       - success
#           undef   - failed
#
sub ConfigureSessionAudit($)
{
    my $args = $_[0];
    my $action = $args->action();
    my $class = $args->class();
    my $regkey = $SESSION_AUDIT_REGKEY;
    my $regval = $SESSION_AUDIT_NAME;
    
    my $reg = CentrifyDC::GP::RegHelper->new($action, $class, $regkey, $regval, undef);
    $reg->load();

    DEBUG_OUT("Configuring DirectAudit session auditing ...");
    
    #
    # Set system value
    #
    
    my $rc = undef;
    if (defined(&CentrifyDC::GP::General::RunCmd))
    {
        $rc = CentrifyDC::GP::General::RunCmd([$DAINFO, "-q", "nss_status"]);
    }
    else
    {
        $rc = CentrifyDC::GP::General::RunCommand("$DAINFO -q nss_status");
    }
    
    if (!defined($rc) || $rc < 0)
    {
        ERROR_OUT("Failed to query the status of DirectAudit session auditing.");
        return undef;
    }
    my $status = ($rc eq '0' ? $SESSION_AUDIT_VALUE_ENABLED : $SESSION_AUDIT_VALUE_DISABLED);
    $reg->set('system', $status);
    
    #
    # Apply new GP setting
    #
    my $apply_group = $reg->getGroupToApply($action);

    if (defined($apply_group) && $apply_group ne "")
    {
        my $value = $reg->get($apply_group);

        if ($value ne $SESSION_AUDIT_VALUE_ENABLED && $value ne $SESSION_AUDIT_VALUE_DISABLED)
        {
            ERROR_OUT("Invalid value [%s] for DirectAudit session auditing", $value);
            return undef;
        }

        my $option = ($value eq $SESSION_AUDIT_VALUE_ENABLED ? '-e' : '-d');
        my $action = ($value eq $SESSION_AUDIT_VALUE_ENABLED ? "enable" : "disable");
        my $output = undef;
        if (defined(&CentrifyDC::GP::General::RunCmd))
        {
            ($rc, $output) = CentrifyDC::GP::General::RunCmd([$DACONTROL, $option]);
        }
        else
        {
            ($rc, $output) = CentrifyDC::GP::General::RunCommand("$DACONTROL $option");
        }
        if ($rc != 0)
        {
            WARN_OUT("Failed to %s DirectAudit session auditing: %s", $action, $rc);
            DEBUG_OUT("Reason: [%s]", $output);
            return undef;
        }

        INFO_OUT("Successfully %sd DirectAudit session auditing.", $action);
    }

    return 1;    
}

#
# Configure audit commands according to GP settings.
#
# We will use the dacontrol command to enable or disable commands to be 
# audited. The dacontrol command needs to provide exit status to indicate if 
# the query/enable/disable operation is successful or failed. Re-enabling the 
# same command should not return error.
#
#   $_[0]:  GP args
#   $_[1]:  command to be audit enabled or disabled
#
#   return: 1       - success
#           undef   - failed
#
sub ConfigureAuditCommand($$)
{
    my $args = $_[0];
    my $command = $_[1];
    
    my $action = $args->action();
    my $class = $args->class();
    my $regkey = $AUDIT_COMMANDS_REGKEY;

    # Since the registry value has been created by the listbox from GP UI, 
    # there is no need to set current and previous values explicitly.
    my $reg_command = CentrifyDC::GP::RegHelper->new($action, $class, $regkey, $command, undef);
    $reg_command->load();

    DEBUG_OUT("Configuring audit command [$command]...");
    
    #
    # Set system value
    #
    
    my $rc = undef;
    if (defined(&CentrifyDC::GP::General::RunCmd))
    {
        $rc = CentrifyDC::GP::General::RunCmd([$DACONTROL, "-q", "-c", $command]);
    }
    else
    {
        $rc = CentrifyDC::GP::General::RunCommand("$DACONTROL -q -c \"$command\"");
    }
    
    if (!defined($rc))
    {
        ERROR_OUT("Failed to query the audit status of command [%s]", $command);
        return undef;
    }
    my $status = ($rc eq '0' ? $AUDIT_COMMANDS_VALUE_ENABLE : $AUDIT_COMMANDS_VALUE_DISABLE);
    $reg_command->set('system', $status);

    #
    # Apply new GP setting
    #

    my $apply_group = $reg_command->getGroupToApply($action);

    if (defined($apply_group) && $apply_group ne "")
    {
        my $action = $reg_command->get($apply_group);
        
        # Any value can be set for listbox in adm file, so we check if value is 
        # invalid here. On the other hand, a drop down list can be added in xml 
        # file so the value can be controlled.
        if ($action ne $AUDIT_COMMANDS_VALUE_ENABLE && $action ne $AUDIT_COMMANDS_VALUE_DISABLE)
        {
            ERROR_OUT("Invalid action [%s] for audit command [%s]", $action, $command);
            return undef;
        }
        
        DEBUG_OUT("%s audit command [%s]", $action, $command);

        my $option = ($action eq $AUDIT_COMMANDS_VALUE_ENABLE ? '-e' : '-d');
        my $output = undef;
        if (defined(&CentrifyDC::GP::General::RunCmd))
        {
            ($rc, $output) = CentrifyDC::GP::General::RunCmd([$DACONTROL, $option, "-c", $command]);
        }
        else
        {
            ($rc, $output) = CentrifyDC::GP::General::RunCommand("$DACONTROL $option -c \"$command\"");
        }
        if ($rc ne '0')
        {
            # The command cannot be audited if it is a link, a shell, or a 
            # wrapper script. Now we will also report these invalid commands.
            WARN_OUT("Failed to %s audit command [%s]: %s", lc($action), $command, $rc);
            DEBUG_OUT("Reason: [%s]", $output);
            return undef;
        }

        INFO_OUT("Successfully %sd audit command [%s]", lc($action), $command);
    }

    return 1;
}

#
# Configure auditing commands. Continue even if failed to configure one or more 
# auditing commands.
#
#   return: 1       - successful
#           undef   - failed
#
sub ConfigureAuditCommands($)
{
    my $args = $_[0];

    my $action = $args->action();
    my $class = $args->class();
    my $regkey = $AUDIT_COMMANDS_REGKEY;

    my $reg = CentrifyDC::GP::RegHelper->new($action, $class, $regkey, undef, undef);
    $reg->load();

    my $ret = 1;
    my $commands = {};

    # Get list of commands from current/previous/local registry
    foreach my $group (qw(current previous local))
    {
        my $command_list = $reg->get($group);
        next if (!defined($command_list));

        foreach my $command (sort keys %$command_list)
        {
            TRACE_OUT("Group: [%s] Command: [%s] Value: [%s]", $group, $command, $command_list->{$command});
            $commands->{$command} = 1;
        }
    }

    # We will NOT proceed if cannot find the dacontrol command. This should be 
    # a corner case since this GP script is shipped with DA
    if (!IsEmpty($commands) && ! -x $DACONTROL)
    {
        ERROR_OUT("Failed to find the dacontrol command on this computer. Audit commands will NOT be configured.");
        return undef;
    }

    # Configure commands
    foreach my $command (sort keys %$commands)
    {
        #
        # Check if command exists. If not, simply ignore the command. Since 
        # admin might add commands from different platforms to this GP, it 
        # would be too harsh if we return failure for this.
        #
        if (! -e $command)
        {
            WARN_OUT("Command [%s] not found. This audit command will NOT be configured. Please make sure the command path is correct.", $command);
            next;
        }
        
        $ret &= ConfigureAuditCommand($args, $command);
    }

    if (!$ret)
    {
        WARN_OUT("Failed to configure all audit commands. Please refer to previous log messages for detail.");
    }

    return $ret;
}

#
# Copy registry policy file to DA local folder
#
sub CopyRegistryPol()
{
    DEBUG_OUT("Copying Registry.pol to /var/centrifyda");

    # Copy the policy file.
    copy(
    $da_gp_policy_file,
    "/var/centrifyda/Registry.pol.new");

    # Atomically rename the new policy into place.
    rename(
    "/var/centrifyda/Registry.pol.new",
    "/var/centrifyda/Registry.pol");
}

#
# Remove registry policy file in DA local folder 
#
sub RemoveRegistryPol()
{
    DEBUG_OUT("Removing /var/centrifyda/Registry.pol");

    # Remove the policy file.
    unlink(
    "/var/centrifyda/Registry.pol");
}

#
# Configure advanced monitoring. 
# Enable or Disable advanced monitoring according to the "Enable advanced monitoring" setting
#
# #   $_[0]:  action  (map/unmap)
#
#   return: 1       - successful
#           0       - no change
#           undef   - failed
#
sub ConfigureAdvancedMonitor($)
{
    my $rc = undef;
    my $output = undef;
    my $status = undef;
    my $retry = 1;

    my $args = $_[0];
    my $action = $args->action();
    my $class = $args->class();

    my $regkey = $AUDIT_ENABLE_REGKEY;
    my $regval = $ADVANCED_MONITOR_ENABLED;
    my $reg = CentrifyDC::GP::RegHelper->new($action, $class, $regkey, $regval, undef);
    $reg->load();

    DEBUG_OUT("Configuring advanced monitoring ...");

    #
    # Set system value
    #

    ($rc, $output) = RunCommand("$DAINFO"); 

    if ($output !~ "DirectAudit advanced monitoring:")
    {
        DEBUG_OUT("Current platform does not support advanced monitoring");
        return 0;
    }

    if ($output =~ "DirectAudit advanced monitoring: Enabled")
    {
        $status = $ADVANCED_MONITOR_VALUE_ENABLED; 
    }
    else 
    {
        $status = $ADVANCED_MONITOR_VALUE_DISABLED; 
    }

    $reg->set('system', $status);

    my $apply_group = $reg->getGroupToApply($action);

    if (defined($apply_group) && $apply_group ne "")
    {
        my $value = $reg->get($apply_group);

        if ($value ne $ADVANCED_MONITOR_VALUE_ENABLED && $value ne $ADVANCED_MONITOR_VALUE_DISABLED && $value ne $ADVANCED_MONITOR_VALUE_NOT_CONF)
        {
            ERROR_OUT("Invalid value [%s] for advanced monitoring.", $value);
            return undef;
        }

        if ($value eq $ADVANCED_MONITOR_VALUE_ENABLED)
        { 
            $rc = RunCommand("$DARELOAD -m");
            DEBUG_OUT("reload settings of advanced monitoring");

            if (!defined($rc) || $rc != 0)
            {
                ERROR_OUT("Cannot reload settings of advanced monitoring");
                return undef;
            }

            # try 5 times if "dacontrol -m" fail 
            # because dacontrol is not stable in some platform
            while($retry < MAX_RETRIES)
            {
               $rc = RunCommand("$DACONTROL -m");
               if ($rc == 0)
               {
                   DEBUG_OUT("advanced monitoring enabled by GP setting");
                   last;
               }
               sleep(1);
               $retry++;
            }

            if (!defined($rc) || $rc != 0)
            {
                ERROR_OUT("Failed to enable advanced monitoring");
                return undef;
            }

            DEBUG_OUT("advanced monitoring enabled by GP setting");
        }
        else
        {
            $rc = RunCommand("$DACONTROL -n");
            DEBUG_OUT("advanced monitoring disabled by GP setting");

            if (!defined($rc) || $rc != 0)
            {
                ERROR_OUT("Failed to disable advanced monitoring");
                return undef;
            }
        }
    }

    return 1;
}

#
# Map function
#
#   return: 1       - successful
#           undef   - failed
#
sub Map($)
{
    my $ret = undef;

    #
    # Note: this mapper script is necessary because it is not safe for
    # DirectAudit to read the policy file directly -- during policy updates
    # the CDC GP processing leaves a window open where the policy file
    # may be missing or incomplete.  See bug 22852.  So, we just copy
    # out the policy to a safe location.
    #
    # (Remark: Other processes should NOT read GP policy file directly. We 
    # should e.g. configure the setting to centrifyda.conf using GP mapper 
    # scripts. Then GP will invoke dareload to signal DA daemon to trigger 
    # corresponding update action.)
    #
    
    if ($is_policy_file_exist)
    {
        CopyRegistryPol();
    }
    else
    {
        RemoveRegistryPol();
    }
    
    $ret = ConfigureAdvancedMonitor($_[0]);
    if (!defined($ret))
    {
        ERROR_OUT("Failed to configure advanced monitoring");
        return undef;
    }
    $ret = ConfigureAuditCommands($_[0]);
    
    $ret = ConfigureSessionAudit($_[0]);

    return $ret;
}

#
# Unmap function
#
#   return: 1       - successful
#           undef   - failed
#
sub UnMap($)
{
    my $ret = undef;
    
    RemoveRegistryPol();

    $ret = ConfigureAdvancedMonitor($_[0]);
    if (!defined($ret))
    {
        ERROR_OUT("Failed to configure advanced monitoring");
        return undef;
    }

    $ret = ConfigureAuditCommands($_[0]);
    
    $ret = ConfigureSessionAudit($_[0]);

    return $ret;
}

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

if (stat("${NEW_DA_GP_POLICY_DIR}/${FILENAME}"))
{
    $da_gp_policy_file = "${NEW_DA_GP_POLICY_DIR}/${FILENAME}";
    $is_policy_file_exist = 1;
}
elsif (stat("${OLD_DA_GP_POLICY_DIR}/${FILENAME}"))
{
    $da_gp_policy_file = "${OLD_DA_GP_POLICY_DIR}/${FILENAME}";
    $is_policy_file_exist = 1;
}

my $args = CentrifyDC::GP::Args->new();
CentrifyDC::GP::Registry::Load($args->user());

my $ret = 0;
$args->isMap() ? ($ret = Map($args)) : ($ret = UnMap($args));
$ret or FATAL_OUT();
