#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2015-2019 Centrify Corporation. All rights reserved.
#
# Machine mapper script to configure Audit Trail settings to Centrify
# DirectControl configuration file.
#
# 1. Audit Trail global targets:
#   'audittrail.targets'
#
# 2. Audit Trail component parameters:
#   'audittrail.Centrify_Suite.Centrify_Commands.*'
#   'audittrail.Centrify_Suite.Centrify_sshd.*'
#   'audittrail.Centrify_Suite.Command.*'
#   'audittrail.Centrify_Suite.DirectControl_UNIX_Agent.*'
#   'audittrail.Centrify_Suite.DirectAudit_UNIX_Agent.*'
#   'audittrail.Centrify_Suite.PAM.*'
#   'audittrail.Centrify_Suite.Trusted_Path.*'
#   'audittrail.Centrify_Suite.Kerberos.*'
#   'audittrail.Centrify_Suite.dzdo.*'
#   'audittrail.Centrify_Suite.dzinfo.*'
#   'audittrail.Centrify_Suite.dzsh.*'
#
# 3. The following parameters only work for Windows side, So we don't need
#    the handle them.
#   'audittrail.Centrify_Suite.Audit_Analyzer.*'
#   'audittrail.Centrify_Suite.Audit_Manager.*'
#   'audittrail.Centrify_Suite.Centrify_Configuration.*'
#   'audittrail.Centrify_Suite.DirectAudit_System_Management.*'
#   'audittrail.Centrify_Suite.DirectAudit_-_Windows.*'
#   'audittrail.Centrify_Suite.DirectAuthorize_-_Windows.*'
#
# Note for developers: DO NOT use centrify_audittrail_settings.in to generate
# script audittail.pl. Since Windows side uses "AuditTrailTargets" and
# "Overrides" as VALUENAME in ADM file for all audit trail components. However,
# policy.pl expects the centrifydc.conf parameter name from VALUENAME.
#
#
#  Map:     Configure settings to centrifydc.conf
#
#  Unmap:   Restore original settings
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;

use CentrifyDC::GP::Mapper;
use CentrifyDC::GP::GPIsolation qw(GetRegKey);

my $file;
my $action;
my $user;
$file = {
    'comment_markers' => [
      '#',
    ],
    'hierarchy_separator' => '.',
    'list_expr' => ', *| +',
    'list_separator' => ', ',
    'lock' => '/etc/centrifydc/centrifydc.conf.lock',
    'match_expr' => [
      '/^\s*([^\s:=]+)[:=]\s*(.*)/',
    ],
    'named_list_separator' => ',',
    'parent_expr' => '^(.*)\.([^\.]+)$',
    'path' => [
      '/etc/centrifydc/centrifydc.conf',
    ],
    'post_action' => [
      'DO_ADRELOAD',
    ],
    'value_map' => {
      # Components' targets
      'audittrail.Centrify_Suite.Centrify_Commands.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Centrify Commands',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Centrify_Commands.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Centrify Commands',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      
      'audittrail.Centrify_Suite.Centrify_sshd.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Centrify sshd',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Centrify_sshd.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Centrify sshd',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Command.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Command',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Command.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Command',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.DirectControl_UNIX_Agent.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.DirectControl UNIX Agent',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.DirectControl_UNIX_Agent.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.DirectControl UNIX Agent',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.DirectAudit_UNIX_Agent.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.DirectAudit UNIX Agent',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.DirectAudit_UNIX_Agent.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.DirectAudit UNIX Agent',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.PAM.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.PAM',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.PAM.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.PAM',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Trusted_Path.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Trusted Path',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Trusted_Path.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Trusted Path',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Kerberos.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Kerberos',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Kerberos.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Kerberos',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },      
      'audittrail.Centrify_Suite.dzdo.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.dzdo',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.dzdo.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.dzdo',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.dzinfo.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.dzinfo',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.dzinfo.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.dzinfo',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.dzsh.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.dzsh',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.dzsh.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.dzsh',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.DirectAudit_Advanced_Monitoring.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.DirectAudit Advanced Monitoring',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.DirectAudit_Advanced_Monitoring.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.DirectAudit Advanced Monitoring',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
        'audittrail.Centrify_Suite.Local_Account_Management.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Local Account Management',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.Local_Account_Management.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.Local Account Management',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },

        'audittrail.Centrify_Suite.MFA.overrides' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.MFA',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'Overrides',
        'value_type' => 'named',
      },
      'audittrail.Centrify_Suite.MFA.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail/Centrify Suite.MFA',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },

      # Global audit trail target
      'audittrail.targets' => {
        'default_data' => '0',
        'reg_class' => 'machine',
        'reg_key' => 'Software/Policies/Centrify/AuditTrail',
        'reg_type' => [
          'REG_DWORD',
        ],
        'reg_value' => 'AuditTrailTargets',
        'value_type' => 'named',
      },
    },
    'write_data' => '$value: $data\n',
};

$action = $ARGV[0];
my $mode = $ARGV[2] ? $ARGV[2] : $ARGV[1];
$user = $ARGV[2] ? $ARGV[1] : undef;

if ($action eq "unmap")
{

    CentrifyDC::GP::Mapper::UnMap($file, $user);
}
else
{
    CentrifyDC::GP::Mapper::Map($file, $user);
}
