#!/bin/sh /usr/share/centrifydc/perl/run

# Copyright (C) 2009-2019 Centrify Corporation. All rights reserved.
# 
# This script retrieves the host name-to-kerberos policy setting from local
# registry and stores it into a local file. (<regdir>/machine/domain_realm).
# This file will be removed on unmap operation or the setting is disabled or
# "not configured" on AD.
# adclient is responsible to merge it to krb5.conf in [domain_realm] section.

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::Config;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::General qw(:debug WriteFile RunCmd);
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);
use Fcntl;
use Fcntl qw(LOCK_EX O_RDWR O_CREAT O_TRUNC);

my $args = CentrifyDC::GP::Args->new('machine');
CentrifyDC::GP::Registry::Load(undef);

my $class = $args->class();
my $group = "current";
my $key = "Software/Microsoft/Windows/Currentversion/Policies/System/Kerberos/domain_realm";
my $key_enabled = "Software/Microsoft/Windows/Currentversion/Policies/System/Kerberos";
my $domain_realm_enabled;
my $dr_values;
my $drfh;

# The domain_realm mapping file store.
my $my_dr_file = $CentrifyDC::Config::properties{"gp.mappers.domain_realm.file"};
if (!defined($my_dr_file))
{
     $my_dr_file = "/etc/centrifydc/domain_realm";
}

if ($args->isMap())
{
    # Read the latest host name-to-realm mappings from local reg.
    (undef, $domain_realm_enabled) = CentrifyDC::GP::Registry::Query("machine", $key_enabled, $group, "domain_realm_Enabled");
    DEBUG_OUT("domain_realm_enabled = $domain_realm_enabled");
     
    if (defined($domain_realm_enabled) and $domain_realm_enabled eq 1)
    {
        # First check if there is any changes in the local vs. reg setting
        # values
        my $anychange = 0;
        my $fh1;
        my $data_in_local_file;

        if (-f "$my_dr_file")
        {
            $data_in_local_file = do {
                open ($fh1, "$my_dr_file") or die "Cannot open file";
                local $/ = undef;
                <$fh1>;
            };
            $data_in_local_file =~ tr /\n//d;
            close($fh1);
        }

        my @data_in_reg = ();
        foreach my $value (CentrifyDC::GP::Registry::Values($class, $key, $group))
        {
            (undef, $dr_values) = CentrifyDC::GP::Registry::Query($class, $key, $group, $value);
            push @data_in_reg, "$value=$dr_values";
            DEBUG_OUT("host name-to-realm mappings: Value Name = $value; Values = $dr_values");
        }
        my $data_in_reg_s = join("", @data_in_reg);

        if ($data_in_local_file eq $data_in_reg_s)
        {
            DEBUG_OUT("There is no changes in domain_realm mapping setting");
        }
        else
        {
            $anychange = 1;
            # Open the domain_realm mapping file store for write.
            my $rc = sysopen($drfh, $my_dr_file, O_RDWR|O_CREAT|O_TRUNC|LOCK_EX, 0600);
            if(!$rc)
            {
                ERROR_OUT("Cannot create/open $my_dr_file");
                die;
            }

            foreach my $value2 (@data_in_reg)
            {
                #Store to file
                print $drfh "$value2\n";
                DEBUG_OUT("Writing to local file: $value2");
            }
            close($drfh);
        }

        if ($anychange eq 1)
        {
            # Run "adflush -t" so that adclient can re-build domain info map and
            # update /etc/krb5.conf with the new changes.
            my $adflush = "$CIMS_USR/sbin/adflush";
            my $cmd = [$adflush, "-t"];
            my ($rc, $out) = RunCmd($cmd);            
        }
    }
    else
    {
        # If host name-to-realm mappings on AD is disabled or not configured.
        # we should remove the file.
        if (-e $my_dr_file)
        {
            unlink($my_dr_file);
        }
    }
}
else
{
    # On unmap we should remove the file store.
    if (-e $my_dr_file)
    {
        unlink($my_dr_file);
    }
}
