#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2008-2019 Centrify Corporation. All rights reserved.
#
# Machine mapper script that configures user/group ignore settings.
#
#  The script get user/group ignore settings from group policy and write
#  them into centrifydc.conf (default) or ignore files.
#
#  If using the CDC version 5.5 or earlier, or using the CDC version 5.6 or later
#  and the populating location is "centrifydcconf" (default), this GP script
#  will stay the old way (like CDC version 5.5 or earlier), that is, modify centrifydc.conf.
#
#  If using the CDC version 5.6 or later and the populating location is "ignorefile",
#  this GP script will modify default ignore files,
#  and make sure parameters nss.<user|group>.ignore in centrifydc.conf
#  will be configured as "file:/etc/centrifydc/<user|group>.ignore".
#  Note that, if users specified list of user/group ignore names,
#  this GP script will populate these user/group names into default ignore files;
#  if users specified a file location, this GP script will restore
#  default ignore files with default list.
#
#  Besides, pam.ignore.users is also updated when updating nss.user.ignore,
#  and make sure their values are same.
#
#  The syntax is one user name per line in ignore files:
#  user_name
#
#
#  Map:     modify centrifydc.conf or ignore files based on group policy setting.
#              Not Configured: restore original setting
#              Enable:         write gp setting into centrifydc.conf or ignore files
#              Disable:        restore original setting
#
#  Unmap:   restore original setting
#
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::RegHelper;
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::Mapper;
use CentrifyDC::GP::General qw(:debug IsEmpty ReadFile WriteFile CreateFile);

# comment line at the beginning of ignore files
my @MARKER = (
    '######################################################################',
    '#        DO NOT EDIT. CentrifyDC group policy generated file.        #',
    '#       Changes to this file can be made through group policy.       #',
    '######################################################################',
);

my @IGNORE_TYPES = qw(user group);

my %NSS_IGNORE = (
    user => "nss.user.ignore",
    group => "nss.group.ignore",
);

my $PAM_IGNORE_USER = "pam.ignore.users";

my %NSS_IGNORE_DEFAULT = (
    user => "nss_user_ignore_default",
    group => "nss_group_ignore_default",
);

my %LOCATION_SWITCH = (
    user => "user_ignore_location_switch",
    group => "group_ignore_location_switch",
);

my %IGNORE_FILE_PATH = (
    user => "/etc/centrifydc/user.ignore",
    group => "/etc/centrifydc/group.ignore",
);

my $file;

my $REGKEY = "Software/Policies/Centrify/CentrifyDC/Settings/Login";

my $CENTRIFYDC_CONF_FLAG     = "centrifydcconf";
my $IGNORE_FILE_FLAG         = "ignorefile";

# hash of registry setting
my $registry_setting;

sub ::DoAdFlush();
sub OldStyleSetting($);
sub NewStyleSetting($);
sub DoIgnoreFileMap($);
sub DoIgnoreFileUnmap();

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# create file for adreload/adflush
#
sub ::DoAdFlush()
{
    CreateFile("/var/centrifydc/reg/do_adflush");
    CreateFile("/var/centrifydc/reg/do_adreload"); 
}

#
# use CDC version 5.5 or earlier map setting for generic map to centrifydc.conf.
#
# note, the map setting of pam.ignore.users is the same as nss.user.ignore.
#
#   $_[0]:  type    (user/group)
#
sub OldStyleSetting($)
{
    my $type = $_[0];

    $file->{value_map}->{$NSS_IGNORE{$type}} = {
      'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} ne "ignorefile"',
      'default_data' => 'file:/etc/centrifydc/user.ignore',
      'reg_class'  => 'machine',
      'reg_key'    => $REGKEY,
      'reg_type'   => [ 'REG_SZ', ],
      'reg_value'  => $NSS_IGNORE{$type},
      'value_type' => 'named',
    };
   
    if ($type eq "user")
    {
        $file->{value_map}->{$PAM_IGNORE_USER} = {
          'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} ne "ignorefile"',
          'default_data' => 'file:/etc/centrifydc/user.ignore',
          'reg_class'  => 'machine',
          'reg_key'    => $REGKEY,
          'reg_type'   => [ 'REG_SZ', ],
          'reg_value'  => $NSS_IGNORE{$type},
          'value_type' => 'named',
        };
    }
}

#
# use CDC version 5.6 or later map setting for generic map to centrifydc.conf,
# taht is, make sure parameters nss.<user|group>.ignore in centrifydc.conf
# will be configured as "file:/etc/centrifydc/<user|group>.ignore".
#
# note, the map setting of pam.ignore.users is the same as nss.user.ignore.
#
#   $_[0]:  type    (user/group)
#
sub NewStyleSetting($)
{
    my $type = $_[0];

    $file->{value_map}->{$NSS_IGNORE{$type}} = {
      'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} eq "ignorefile"',
      'data_value' => $NSS_IGNORE_DEFAULT{$type},
      'default_data' => 'file:/etc/centrifydc/user.ignore',
      'reg_class' => 'machine',
      'reg_key' => $REGKEY,
      'reg_type' => [ 'REG_SZ', ],
      'reg_value' => $NSS_IGNORE{$type},
      'value_type' => 'named',
    };
   
    if ($type eq "user")
    {
        $file->{value_map}->{$PAM_IGNORE_USER} = {
          'active' => '$value_map->{' . $LOCATION_SWITCH{$type} . '}{reg_data} eq "ignorefile"',
          'data_value' => $NSS_IGNORE_DEFAULT{$type},
          'default_data' => 'file:/etc/centrifydc/user.ignore',
          'reg_class' => 'machine',
          'reg_key' => $REGKEY,
          'reg_type' => [ 'REG_SZ', ],
          'reg_value' => $NSS_IGNORE{$type},
          'value_type' => 'named',
        };
    }
}

#
# populate user/group ignore names into default ignore files.
#
#  - if using the CDC version 5.6 or later and the populating location is "ignorefile".
#    - if users specified a file location, restore default ignore files with default list.
#    - if users specified list of user/group ignore names, populate these user/group names into default ignore files.
#  - if using the CDC version 5.5 or earlier, or using the CDC version 5.6 or later and the populating location is "centrifydcconf".
#    - restore default ignore files with default list.
#
#   $_[0]:  action  (map/unmap)
#
sub DoIgnoreFileMap($)
{
    my $action = $_[0];
    my $changed = 0;

    foreach my $type (@IGNORE_TYPES)
    {
        my $is_new_style = ($registry_setting->{$type}->{location_cur} eq $IGNORE_FILE_FLAG);
        DEBUG_OUT("Do %s style %s.ignore map.", $is_new_style ? "new" : "old", $type);

        my $reg_ignores = CentrifyDC::GP::RegHelper->new($action, 'machine', $REGKEY, $NSS_IGNORE{$type}, undef);
        $reg_ignores or return undef;
        $reg_ignores->load();
        my $setting = $reg_ignores->get("current");

        if ($is_new_style)
        {
            if (!-e $IGNORE_FILE_PATH{$type})
            {
            	# almost never happens.
                DEBUG_OUT("$IGNORE_FILE_PATH{$type} doesn't exist.");
            }
            else
            {
                my $ignore_file_bk = $IGNORE_FILE_PATH{$type} . ".bak";
                if (!-e $ignore_file_bk)
                {
                    DEBUG_OUT("Backup $IGNORE_FILE_PATH{$type} to $ignore_file_bk.");
                    link($IGNORE_FILE_PATH{$type}, $ignore_file_bk) || FATAL_OUT $!;
                }
            }
           
            if ($setting =~ /^file:*/)
            {
                DEBUG_OUT("User specifies a file location, restore default %s ignore file setting.", $type);
                my $ignore_file_bk = $IGNORE_FILE_PATH{$type} . ".bak";
                if (-e $ignore_file_bk)
                {
                    if (-e $IGNORE_FILE_PATH{$type})
                    {
                        unlink($IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
                    }
                    link($ignore_file_bk, $IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
                    unlink($ignore_file_bk) || FATAL_OUT $!;
                }
                next;
            }

            DEBUG_OUT("User specifies some %s names, populate to %s.ignore file.", $type, $type);

            $setting =~ s/^\s+//;
            $setting =~ s/\s+$//;
            $setting =~ s/\s+/\n/g;
            my $str = join("\n", @MARKER);
            $str .= "\n" . $setting . "\n";
            my $sys_string = ReadFile($IGNORE_FILE_PATH{$type});
            if (!defined($sys_string) || $sys_string ne $str)
            {
                DEBUG_OUT("Write %s ignore file " . $IGNORE_FILE_PATH{$type}, $type);
                unlink($IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
                my $rc = WriteFile($IGNORE_FILE_PATH{$type}, $str, {'mode' => 0644});
                if (! $rc)
                {
                    ERROR_OUT("Cannot write user ignore file " . $IGNORE_FILE_PATH{$type});
                    return;
                }
                $changed = 1;
            }
        }
        else
        {
            my $ignore_file_bk = $IGNORE_FILE_PATH{$type} . ".bak";
            if (-e $ignore_file_bk)
            {
                DEBUG_OUT("Restore %s ignore file setting.", $type);
                if (-e $IGNORE_FILE_PATH{$type})
                {
                    unlink($IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
                }
                link($ignore_file_bk, $IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
                unlink($ignore_file_bk) || FATAL_OUT $!;
                $changed = 1;
            }
        }
    }

    if ($changed)
    {
        DEBUG_OUT("User/Group ignore file setting changed. Need to do adflush.");
        DoAdFlush();
    }
}

#
# restore default ignore files.
#
#   $_[0]:  action  (map/unmap)
#
sub DoIgnoreFileUnmap()
{
    foreach my $type (@IGNORE_TYPES)
    {
        DEBUG_OUT("Restore %s ignore file setting.", $type);
        my $ignore_file_bk = $IGNORE_FILE_PATH{$type} . ".bak";
        if (-e $ignore_file_bk)
        {
            if (-e $IGNORE_FILE_PATH{$type})
            {
                unlink($IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
            }
            link($ignore_file_bk, $IGNORE_FILE_PATH{$type}) || FATAL_OUT $!;
            unlink($ignore_file_bk) || FATAL_OUT $!;
        }
    }
}

# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

$file = {
    'comment_markers' => [
      '#',
    ],
    'hierarchy_separator' => '.',
    'list_expr' => ', *| +',
    'list_separator' => ', ',
    'match_expr' => [
      '/^\s*([^\s:=]+)[:=]\s*(.*)/',
    ],
    'parent_expr' => '^(.*)\.([^\.]+)$',
    'path' => [
      '/etc/centrifydc/centrifydc.conf',
    ],
    'lock' => '/etc/centrifydc/centrifydc.conf.lock',
    'post_command' => '::DoAdFlush()',
    'value_map' => {
      'user_ignore_location_switch' => {
        'active' => '0',
        'reg_class' => 'machine',
        'reg_key' => "$REGKEY",
        'reg_type' => [
          'REG_SZ',
        ],
        'reg_value' => 'user_ignore_location_switch',
        'value_type' => 'named',
      },
      'group_ignore_location_switch' => {
        'active' => '0',
        'reg_class' => 'machine',
        'reg_key' => "$REGKEY",
        'reg_type' => [
          'REG_SZ',
        ],
        'reg_value' => 'group_ignore_location_switch',
        'value_type' => 'named',
      },
    },
    'write_data' => '$value: $data\n',
};

$registry_setting = {
    'user' => {
        'location_cur' => "$CENTRIFYDC_CONF_FLAG",
        'location_pre' => "$CENTRIFYDC_CONF_FLAG",
    },
    'group' => {
        'location_cur' => "$CENTRIFYDC_CONF_FLAG",
        'location_pre' => "$CENTRIFYDC_CONF_FLAG",
    },
};

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new('machine');

CentrifyDC::GP::Registry::Load(undef);

foreach my $type (@IGNORE_TYPES)
{
    my $cur = (CentrifyDC::GP::Registry::Query($args->class(), $REGKEY, "current", $LOCATION_SWITCH{$type}))[1];
    my $pre = (CentrifyDC::GP::Registry::Query($args->class(), $REGKEY, "previous", $LOCATION_SWITCH{$type}))[1];
    if (defined $cur)
    {
        $registry_setting->{$type}->{location_cur} = $cur;
    }
    if (defined $pre)
    {
        $registry_setting->{$type}->{location_pre} = $pre;
    }
}


if ($args->isMap())
{
    DoIgnoreFileMap($args->action());
    
    # Since the map setting ($file) for the same parameters (nss.<user|group>.ignore and pam.ignore.users)
    # are different under different populating locations or different CDC versions,
    # so after user switches the populating location, we must perform the previous setting Map()
    # to clear the previous configuration.
    my $do_pre_map = 0;
    foreach my $type (@IGNORE_TYPES)
    {
        if ($registry_setting->{$type}->{location_pre} ne $registry_setting->{$type}->{location_cur})
        {
            if ($registry_setting->{$type}->{location_pre} eq $IGNORE_FILE_FLAG)
            {
                NewStyleSetting($type);
            }
            else
            {
                OldStyleSetting($type);
            }
            $do_pre_map = 1;
        }
    }
    if ($do_pre_map)
    {
        DEBUG_OUT("The CDC version changed, or user switched the populating location, need to do previous setting map.");
        CentrifyDC::GP::Mapper::Map($file, $args->user());
    }

    foreach my $type (@IGNORE_TYPES)
    {
        if ($registry_setting->{$type}->{location_cur} eq $IGNORE_FILE_FLAG)
        {
            NewStyleSetting($type);
        }
        else
        {
            OldStyleSetting($type);
        }
    }
    DEBUG_OUT("Do generic map.");
    CentrifyDC::GP::Mapper::Map($file, $args->user());
}
else
{
    DoIgnoreFileUnmap();

    foreach my $type (@IGNORE_TYPES)
    {
        if ($registry_setting->{$type}->{location_cur} eq $IGNORE_FILE_FLAG)
        {
            NewStyleSetting($type);
        }
        else
        {
            OldStyleSetting($type);
        }
    }
    DEBUG_OUT("Do generic unmap.");
    CentrifyDC::GP::Mapper::UnMap($file, $args->user());
}

