##############################################################################
#
# Copyright (C) 2004-2019 Centrify Corporation. All rights reserved.
#
# Centrify DirectControl mapper script Directory Access module.
#
# This module uses adquery and ldapsearch to get user info.
#
##############################################################################

use strict;

package CentrifyDC::GP::DirectoryAccess;
my $VERSION = '1.0';
require 5.000;

use vars qw(@ISA @EXPORT_OK);

require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(GetAttribute GetQueryInfo GetHomeDirectoryProperty GetHomeURL);

use CentrifyDC::GP::General qw(:debug IsEmpty RunCmd);
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);

my $ADQUERY     = "$CIMS_USR/bin/adquery";
my $LDAPSEARCH  = "$CIMS_INSTALL/bin/ldapsearch";

sub GetQueryInfo($$);
sub GetAttribute($$$$$;$);
sub GetHomeDirectoryProperty($);
sub GetHomeURL($);
sub _TrimURL($);



#
# Use adquery to get all information of specified user/group
#
#   $_[0]: type (user/group)
#   $_[1]: user/group name
#
#   return: 
#       hash reference  - user/group information
#                         => {
#                            samAccountName => user,
#                            canonicalName  => domain/OU/user,
#                            ...
#       undef           - failed
#
sub GetQueryInfo($$)
{
    my ($type, $name) = @_;

    if (! defined($type) || ($type ne 'user' && $type ne 'group'))
    {
        ERROR_OUT("Cannot query: incorrect type");
        return;
    }

    if (! defined($name))
    {
        ERROR_OUT("Cannot query: name undefined");
        return;
    }

    TRACE_OUT("Get all information of $type [$name]");

    if (! -f $ADQUERY or ! -x $ADQUERY)
    {
        ERROR_OUT("Cannot find adquery");
        return;
    }

    my ($ret, $output) = RunCmd([$ADQUERY, $type, "-A", $name]);

    if (! defined($ret) or $ret ne 0 or ! defined($output))
    {
        ERROR_OUT("adquery failed: $output");
        return;
    }

    chomp $output;

    my $info;
    my @arr = split('\n', $output);

    foreach my $line (@arr)
    {
        chomp($line);
        if ($line =~ m/([^:]+):(.*)$/)
        {
            $info->{$1} = $2;
        }
    }

    TRACE_OUT("=== begin $type $name info ===");
    foreach my $key (sort keys %$info)
    {
        my $val = $info->{$key};
        TRACE_OUT("| %-20s: [%s]", $key, $val);
    }
    TRACE_OUT("=== end $type $name info ===");

    return $info;
}

#
# Use ldapsearch to get object attribute
#
#   $_[0]: user (if specified, run ldapsearch as user)
#   $_[1]: options
#   $_[2]: ldapuri
#   $_[3]: searchbase
#   $_[4]: filter
#   $_[5]: attribute (optional. if ignored, return the whole output of ldapsearch)
#
#   return: 
#       $1: 
#           1       - successful
#           undef   - failed
#       $2:
#           string  - attribute setting
#
sub GetAttribute($$$$$;$)
{
    my ($user, $options, $ldapuri, $searchbase, $filter, $attribute) = @_;

    defined($options) or $options = '';
    defined($ldapuri) or return;
    defined($searchbase) or return;
    defined($filter) or $filter = '';
    defined($attribute) or $attribute = '';

    if ($options eq '')
    {
        if (defined($user))
        {
            $options = "-r -Q -LLL";
        }
        else
        {
            $options = "-r -m -Q -LLL";
        }
    }

    TRACE_OUT("Get attribute:\n options: [$options]\n ldapuri: [$ldapuri]\nsearchbase: [$searchbase]\nfilter: [$filter]\n Attribute: [$attribute]");

    if (! -f $LDAPSEARCH or ! -x $LDAPSEARCH)
    {
        ERROR_OUT("Cannot find ldapsearch");
        return;
    }

    my $command = '';

    # if current user is different from the specified user, use su to run command
    my $su = 0;
    if (defined($user))
    {
        my $cur_user = (getpwuid($>))[0];
        if ($cur_user ne $user)
        {
            $su = 1;
        }
    }

    ($su) and $command .= "su - '$user' -c \" ";
    $command .= "'$LDAPSEARCH' $options -H '$ldapuri' -b '$searchbase'";
    ($filter ne '') and $command .= " '($filter)'";
    ($attribute ne '') and $command .= " '$attribute'";
    ($su) and $command .= "\"";

    my ($ret, $output) = RunCmd($command);
    if (! defined($ret) or $ret ne 0)
    {
        ERROR_OUT("ldapsearch failed");
        return;
    }

    # remove all the junks and get attribute
    if ($attribute ne '')
    {
        if ($output =~ m/^$attribute:/m)
        {
            $output =~ m/^$attribute:\s*(.*)$/m;
            $output = $1;
        }
        else
        {
            $output = undef;
        }
    }

    return (1, $output);
}


#
# trim share url, convert \ to /, replace space with %20, remove trailing /
#
#   $_[0]:  share url
#
#   return: string  - trimmed share rul
#           undef   - failed
#
sub _TrimURL($)
{
    my $url = $_[0];
    defined($url) or return undef;

    $url =~ s|\\|/|g;   # replace \ with /
    $url =~ s|/+$||g;   # remove trailing /

    return $url;
}

#
# Get user's homeDirectory property
# Need to notice the following:
#  1. user may not be in the joined domain, for example in a trusted domain
#  2. for the above reason, need to use user's credential to do ldapsearch
#
#   $_[0]:  username
#
#   return: string  - homeDirectory property. empty string means this property
#                     doesn't exist.
#           undef   - failed
#
sub GetHomeDirectoryProperty($)
{
    my $user = $_[0];

    defined($user) or return undef;

    TRACE_OUT("Get homeDirectory property of user [$user]");

    my $userinfo = CentrifyDC::GP::DirectoryAccess::GetQueryInfo('user', $user);

    if (! defined($userinfo))
    {
        ERROR_OUT("Cannot query user $user");
        return undef;
    }

    # need to get the following info:
    #  - canonicalName (to calculate ldapuri and searchbase)
    #  - userPrincipalName
    my $user_cn = $userinfo->{canonicalName};
    my $user_upn = $userinfo->{userPrincipalName};
    my $user_sam = $userinfo->{samAccountName};
    if (! defined($user_cn))
    {
        ERROR_OUT("Cannot get canonicalName of user [$user]");
        return undef;
    }

    if (! defined($user_upn) && ! defined($user_sam))
    {
        ERROR_OUT("Cannot get userPrincipalName and samAccountName of user [$user], exiting.");
        return undef;
    }

    $user_cn =~ m|^([^/]+)|;
    my $user_domain = $1;
    if (! defined($user_domain) || $user_domain eq '')
    {
        ERROR_OUT("Cannot extract user domain from canonicalName");
        return undef;
    }
    my $user_domain_dn = $user_domain;
    $user_domain_dn = "DC=" . $user_domain_dn;
    $user_domain_dn =~ s/\./,DC=/g;

    my $ldapuri = "ldap://$user_domain";
    my $filter = "&(objectclass=user)(userPrincipalName=$user_upn)";
    if (! defined($user_upn))
    {
        DEBUG_OUT("Cannot get the userPrincipalName of the user [$user]. Will use samAccountName [$user_sam] instead.");
        $filter = "&(objectclass=user)(samAccountName=$user_sam)";
    }

    my ($ret, $homeDirectory) = CentrifyDC::GP::DirectoryAccess::GetAttribute($user, undef, $ldapuri, $user_domain_dn, $filter, 'homeDirectory');
    if (! $ret)
    {
        ERROR_OUT("Cannot get user's local home directory");
        return undef;
    }

    defined($homeDirectory) or $homeDirectory = '';

    TRACE_OUT("homeDirectory property of user $user: [%s]", $homeDirectory);

    return $homeDirectory;
}

#
# Get user's Windows home URL
#
#   $_[0]:  username
#
#   return: string  - the homeDirectory URL (in smb://server/share format)
#                     empty string means user has no homeDirectory
#           undef   - failed
#
sub GetHomeURL($)
{
    my $user = $_[0];

    defined($user) or return undef;

    my $homedir = GetHomeDirectoryProperty($user);
    defined($homedir) or return undef;

    my $homeURL = '';

    if ($homedir ne '')
    {
        $homedir = _TrimURL($homedir);
        $homeURL = "smb:" . $homedir;
        DEBUG_OUT("User's Windows home URL: [%s]", $homeURL);
    }
    else
    {
        DEBUG_OUT("User's Windows home URL is empty");
    }

    return $homeURL;
}

1;
