#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2019 Centrify Corporation. All rights reserved.
#
# a Perl package supplying some basic functions to modify pam conf files
# The functions here are used by other /cli/confgurefiles perl scripts
#
##############################################################################

package autoedit_pam_conf;

BEGIN
{
    use lib "/usr/share/centrifydc/scripts";
    use lib "/usr/share/centrifydc/perl";
}

use strict;
use warnings;
use File::Copy qw(copy);

use autoedit_common_utils qw(insert_cfg_lines);
use autoedit_file_lock;
use CentrifyDC::GP::General qw(:debug :system);

use base 'Exporter';
our @EXPORT_OK = qw(process_pam_conf_services copy_pam_conf_service process_pam_conf);

# Process pam.conf services
#
# add our code ($newLines) for each service.
# May have multiple modules for one service
# (e.g. "auth", "account", "session", "password")
# We add our new code seperately for each module.
#
# but not all pam lib like our code to be above them,
# if found any one, add our code behind it,
# and our code may be different (use $newLines2)
#
# $cfgLines         the content of file, which is from <FILE>
# $newLines         our new code to add before $service_module
# $service_module   the target service and module, 
#                   update their code to use Centrify pam
#                   It is a Ref of hash, the key is "service module".
# $libs_excluded    the pam libs, which do not like our code to be above them
#                   It is a Ref of array
# $newLines2        our new code to add behind $service_module if found $libs_excluded
#
sub process_pam_conf_services
{
    my ($cfgLines, $newLines, $service_module, $libs_excluded, $newLines2) = @_;

    # analyse our new code
    # put them into hash by different module name
    my %hash = ();
    my %hash2 = ();
    foreach my $module ("auth", "account", "session", "password")
    {
        my @lines  = grep /\b$module\b/, @$newLines;
        my @lines2 = grep /\b$module\b/, @$newLines2;
        $hash{$module}  = scalar(@lines)  ? [@lines]  : undef;
        $hash2{$module} = scalar(@lines2) ? [@lines2] : undef;
    }

    foreach my $pair (@$service_module)
    {
        my @items = split /\s+/, $pair;
        my $service = $items[0];
        my $module = $items[1];

        # prepare our new code
        # indicate the service name in Centrifydc code
        my $new_Lines = undef;
        if ($hash{$module})
        {
            my @lines = @{$hash{$module}};
            $new_Lines = [grep s/(.*)/$service\t$1/s, @lines];
        }

        my $new_Lines2 = undef;
        if ($hash2{$module})
        {
            my @lines = @{$hash2{$module}};
            $new_Lines2 = [grep s/(.*)/$service\t$1/s, @lines];
        }

        insert_cfg_lines($cfgLines, $new_Lines, $service,
                $module, $libs_excluded->{$module}, $new_Lines2);
    }
}

# copy pam code from $src service for $dst service
# if $dst has existed, do nothing
#
# $cfgLines         the content of file, which is from <FILE>
# $src              the name of source service
# $dst              the name of destination service
# return            0 if meet anything wrong
#                   1 if succeed
sub copy_pam_conf_service
{
    my ($cfgLines, $src, $dst) = @_;

    # return 1, if found $dst
    # it must have been updated.
    if (grep /^[^#]*\b$dst\b/, @$cfgLines)
    {
        return 1;
    }

    my @srcLines = grep /^\s*$src\s+/, @$cfgLines;

    # src does not exist
    if (scalar(@srcLines) == 0)
    {
        return 0;
    }

    # src is prohibited
    if (grep /\bpam_prohibit\b/, @srcLines)
    {
        return 0;
    }

    # copy the code of $src to $dst
    # don't use "grep s/$src/$dst/g" here !
    # we may meet some lib name containing $src (e.g. pam_dzdo ?)
    @srcLines = grep s/^\s*$src(\s+.*)/$dst$1/, @srcLines;

    # for dzdo, add xxx_first_pass
    # do that only if the pam lib is pam_unix
    if ($dst eq "dzdo")
    {
        my $libs = "(pam_unix.so.1|pam_unix.1|pam_unix.so)";
        grep s/(.*\bauth\b.*$libs.*)/$1 use_first_pass/, @srcLines;
        grep s/(.*\bpassword\b.*$libs.*)/$1 try_first_pass/, @srcLines;
    }

    unshift @$cfgLines, @srcLines;
    return 1;
}

# Process pam.conf file
# go through all services in pam.conf
# and add Centrifydc code for each of them
#
# $cdc_pam_lib      the name of Centrifydc pam lib
# $libs_excluded    the pam libs, which do not like our code to be above them
#                   It is a Ref of array
# $cfgName          the full path of pam.conf
# $cdc_long_comment Centrifydc comment added before its code
# $newLines         our new code to add before $service_module
# $newLines2        our new code to add behind $service_module if found $libs_excluded
# return            undef if have anything wrong
#                   1 if succeed
sub process_pam_conf
{
    my ($cdc_pam_lib, $libs_excluded, $cfgName, $cdc_long_comment,
        $newLines, $newLines2) = @_;
    my $cfgHandler;
    my $tmpName;
    my $tmpHandler;

    my $flock = autoedit_file_lock::new("$cfgName");
    if (!defined $flock)
    {
        DEBUG_OUT("cannot lock file $cfgName");
        return undef;
    }

    $cfgName = TraverseSymLink($cfgName);
    if (!defined $cfgName)
    {
        ERROR_OUT("meet error when check symlink $cfgName");
        return undef;
    }

    # to make logic simple, read all lines here
    my $cfgLines = ReadFile("$cfgName");
    if (!defined $cfgLines)
    {
        ERROR_OUT("cannot read $cfgName");
        return undef;
    }
    my @cfgLines = split "\n", $cfgLines;
    @cfgLines = grep {$_ .= "\n"} @cfgLines;

    # return 1, if found centrify pam lib
    # it must have been updated.
    if(grep /^[^#]*\b$cdc_pam_lib\b/, @cfgLines)
    {
        return 1;
    }

    # collect all "Service Module" pair
    my %service_module_hash = ();
    my @modules = ("auth", "account", "session", "password");
    foreach my $line (@cfgLines)
    {
        #skip comment
        if ($line =~ /^\s*#/)
        {
            next;
        }
        my @items = split /\s+/, $line;
        if (defined $items[1] && grep /\b$items[1]\b/, @modules)
        {
            $service_module_hash{"$items[0] $items[1]"} = 1;
        }
    }

    # some services need keyword "try_first_pass"
    # remember their name, and process them later
    my @passwd_services = ();
    foreach ("passwd auth")
    {
        if (exists $service_module_hash{$_})
        {
            delete $service_module_hash{$_};
            push @passwd_services, $_;
        }
    }

    # some services need keyword "enable_dzpamgate"
    # remember their name, and process them later
    my @su_services = ();
    foreach ("su auth")
    {
        if (exists $service_module_hash{$_})
        {
            delete $service_module_hash{$_};
            push @su_services, $_;
        }
    }

    # create and add pam settings for dzdo
    {
        # copy pam code from OTHER (mainly for HPUX)
        my $rc = copy_pam_conf_service(\@cfgLines, "OTHER", "dzdo");

        if (!$rc)
        {
            $rc = copy_pam_conf_service(\@cfgLines, "other", "dzdo");
        }

        if (!$rc)
        {
            # "OTHER" may be prohibited on AIX
            # try to copy pam code from "login"
            $rc = copy_pam_conf_service(\@cfgLines, "login", "dzdo");
        }

        if ($rc)
        {
            # will add Centrifydc code for dzdo later
            $service_module_hash{"dzdo auth"} = 1;
            $service_module_hash{"dzdo account"} = 1;
            $service_module_hash{"dzdo session"} = 1;
            $service_module_hash{"dzdo password"} = 1;

            # copy the code of dzdo to dzdo-mfa and dzdo-ad-mfa
            $rc = copy_pam_conf_service(\@cfgLines, "dzdo", "dzdo-mfa");
            if ($rc)
            {
                $service_module_hash{"dzdo-mfa auth"} = 1;
                $service_module_hash{"dzdo-mfa account"} = 1;
                $service_module_hash{"dzdo-mfa session"} = 1;
                $service_module_hash{"dzdo-mfa password"} = 1;            
            }

            $rc = copy_pam_conf_service(\@cfgLines, "dzdo", "dzdo-ad-mfa");  
            if ($rc)
            {
                $service_module_hash{"dzdo-ad-mfa auth"} = 1;
                $service_module_hash{"dzdo-ad-mfa account"} = 1;
                $service_module_hash{"dzdo-ad-mfa session"} = 1;
                $service_module_hash{"dzdo-ad-mfa password"} = 1;            
            }            
        }
    }
    
    # create and add pam settings for sshd-sso-mfa.
    {
        # copy pam code from sshd
        my $rc = copy_pam_conf_service(\@cfgLines, "sshd", "sshd-sso-mfa");

        if ($rc)
        {
            $service_module_hash{"sshd-sso-mfa auth"} = 1;
            $service_module_hash{"sshd-sso-mfa account"} = 1;
            $service_module_hash{"sshd-sso-mfa session"} = 1;
            $service_module_hash{"sshd-sso-mfa password"} = 1;            
        }
    }

    # get down to some serious work
    my @service_module = keys %service_module_hash;
    if (scalar(@service_module))
    {
        # add our code
        # then move our code to the top of file
        # not necessary, but it looks better to have just one comment at the top.
        my @topLines = @cfgLines;
        process_pam_conf_services(\@topLines, $newLines, \@service_module, $libs_excluded, undef);
        @topLines = grep /^[^#]*\b$cdc_pam_lib\b/, @topLines;
        if (scalar(@topLines))
        {
            @cfgLines = ($cdc_long_comment, @topLines, @cfgLines);
        }
        # append our code after some libes who don't want our code bafore them
        process_pam_conf_services(\@cfgLines, undef, \@service_module, $libs_excluded, $newLines2);
    }

    # deal with passwd services, add try_first_pass
    if (scalar(@passwd_services))
    {
        my @lines = @$newLines;
        @lines = grep {$_ !~ /\bauth\b/} @lines;

        unshift @lines, (
            "auth\tsufficient\t$cdc_pam_lib try_first_pass\n",
            "auth\trequisite\t$cdc_pam_lib deny\n"
            );

        process_pam_conf_services(\@cfgLines, \@lines, \@passwd_services, $libs_excluded, \@lines);
    }

    # deal with su service, add enable_dzpamgate
    if (scalar(@su_services))
    {
        my @lines = @$newLines;
        @lines = grep {$_ !~ /\bauth\b/} @lines;

        unshift @lines, (
            "auth\tsufficient\t$cdc_pam_lib enable_dzpamgate\n",
            "auth\trequisite\t$cdc_pam_lib deny\n",
            );

        process_pam_conf_services(\@cfgLines, \@lines, \@su_services, $libs_excluded, \@lines);
    }

    $cfgLines = join "", @cfgLines;
    if (! copy($cfgName, "$cfgName.pre_cdc"))
    {
        ERROR_OUT("Cannot backup $cfgName: $!");
        return undef;
    }

    if (!WriteFile($cfgName, $cfgLines))
    {
        ERROR_OUT("Cannot write $cfgName");
        return undef;
    }

    return 1;
}

1;
