#!/bin/sh /usr/share/centrifydc/perl/run.cda

##############################################################################
#
# Copyright (C) 2014-2019 Centrify Corporation. All rights reserved.
#
# This script is used to modify apparmor configure files
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts/da";
use CentrifyDA::Include;
use Getopt::Long qw(GetOptions);

use autoedit_common_utils qw(loadSetting getValueList);
use autoedit_constants qw($CDA_Version);
use autoedit_file_lock;
use CentrifyDA::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify apparmor configure files
    usage:
    --cfg_files         the configure files full path
    --do_load_cdc_conf  set it if want to load the configure files path
                        from centrifydc.conf
    --do_revert         set it if want to revert our changes
END
    exit 1;
}

# process apparmor file
# all work here is to add "#include <abstractions/centrifyda>"
# $cfgName      the full path of config file
# $do_revert    do revert
# return        0 if meet error
#               1 if succeed
sub do_process_file
{
    my ($cfgName, $do_revert) = @_;

    my $flock = autoedit_file_lock::new("$cfgName");
    if (!defined $flock)
    {
        warn "cannot lock file $cfgName";
        return 0;
    }

    $cfgName = TraverseSymLink($cfgName);
    if (!defined $cfgName)
    {
        DEBUG_OUT("meet error when check symlink $cfgName");
        return 0;
    }

    my $cfgLines = ReadFile("$cfgName");
    if (!defined $cfgLines)
    {
        DEBUG_OUT("cannot read $cfgName");
        return 0;
    }

    my $newLines = << "END";
  # lines inserted by Centrify Direct Audit $CDA_Version
  # CentrifyDA name service and authentication
  #include <abstractions/centrifyda>
END

    if ($do_revert)
    {
        # remove CentrifyDA additions
        $cfgLines =~ s/ *#.*centrifyda.*\n//img;
    }
    elsif ($cfgLines =~ /<abstractions\/centrifyda>/)
    {
        return 1;
    }
    else
    {
        if ($cfgLines !~ /\n$/)
        {
            # not end with "\n" ?
            $cfgLines .= "\n";
        }
        $cfgLines .= $newLines;
    }

    my $backup = "$cfgName.pre_cdc";
    if (!$do_revert)
    {
        if (! CopyFile($cfgName, $backup))
        {
            ERROR_OUT("Cannot backup $cfgName");
            return 0;
        }
    }
    else
    {
        # install.sh and rpm/pkg scripts may override config files with
        # the backup files if found them exist.
        # To avoid affecting these code logic, we'd better delete 
        # backup file after "do_revert"
        if (-e $backup)
        {
            unlink $backup;
        }
    }

    if (!WriteFile($cfgName, $cfgLines))
    {
        ERROR_OUT("Cannot write $cfgName");
        return 0;
    }

    return 1;
}

my $cfg_files;
my $do_load_cdc_conf;
my $do_revert;

GetOptions( "cfg_files=s" => \$cfg_files,
            "do_load_cdc_conf" => \$do_load_cdc_conf,
            "do_revert" => \$do_revert,
            ) or usage;

if (defined $cfg_files)
{
    $cfg_files = [split /[,\s]+/, $cfg_files];
}
else
{
    my $cfg_dir;
    my $dir = "/etc/apparmor.d/abstractions";

    if (-e "$dir/authentication")
    {
        $cfg_dir = $dir;
    }

    if (!$cfg_dir)
    {
        DEBUG_OUT("cannot find apparmor configuration files");
        exit 1;
    }

    if ($do_load_cdc_conf)
    {
        $cfg_files = getValueList("apparmor.config.profiles");
    }

    if (!$cfg_files)
    {
        $cfg_files = ["authentication", "nameservice"];
    }

    # need full path
    grep {$_ = "$cfg_dir/$_"} @$cfg_files;
}

# Check if the init script exists first.
# If it does not exist, check with "systemctl list-unit-files".
# This way we do not need to check the OS nor the OS version.

my $condrestart_command = undef;

# Looking for init.d script
my $init_script = undef;
foreach my $file ("/etc/init.d/boot.apparmor",
                  "/etc/init.d/apparmor")
{
    if (-x $file)
    {
        $init_script = $file;
        last;
    }
}

# Looking for systemd service
my $systemctl = undef;
if (!defined $init_script)
{
    DEBUG_OUT("cannot find apparmor init.d script");

    if (! -x "/bin/systemctl")
    {
        DEBUG_OUT("cannot find systemd");
        exit 1;
    }

    $systemctl = "/bin/systemctl";

    # Looking for apparmor.service
    my ($ret, $output) = RunCommand("$systemctl list-unit-files | /bin/grep -i apparmor.service > /dev/null");
    if (!$ret)
    {
        $condrestart_command = "$systemctl try-restart apparmor.service";
    }
    else
    {
        DEBUG_OUT("cannot find apparmor service unit");
        exit 1;
    }
}

# Process apparmor files
foreach my $file (@$cfg_files)
{
    if (-e $file)
    {
        my $ret;
        $ret = do_process_file($file, $do_revert);
        if (!$ret)
        {
            exit 1;
        }
    }
}

# Restart apparmor only if it's running (active)
if ($condrestart_command)
{
    my ($ret, $output) = RunCommand($condrestart_command);
    DEBUG_OUT($output);
    if ($ret != 0)
    {
        exit $ret;
    }
}
else
{
    # init.d scripts may don't have condrestart or try-restart options on some platforms.
    # and try-restart option doesn't work as our expect on SuSE11. So use a common approch instead.
    my ($ret, $output) = RunCommand("$init_script status");
    if (!$ret)
    {
        ($ret, $output) = RunCommand("$init_script restart");
        DEBUG_OUT($output);
        if ($ret != 0)
        {
            exit $ret;
        }
    }
}

exit 0;
