#!/bin/sh /usr/share/centrifydc/perl/run.cda

##############################################################################
#
# Copyright (C) 2014-2019 Centrify Corporation. All rights reserved.
#
# This script is used to modify methods.cfg
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error (general error)
#       2 if target file is empty
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts/da";
use CentrifyDA::Include;
use Getopt::Long qw(GetOptions);

use autoedit_file_lock;
use CentrifyDA::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify methods.cfg
    --lib_name              the name of centrifydc or centrifyda lib
    --cfg_file              the full path of methods.cfg
    --is_method_hooked      check if "CENTRIFYDALOCAL" found in /usr/lib/security/methods.cfg
                            this option has to be used together with "--lib_name centrifyda"
    --is_method_normal      check if the methods config is as our expectation in /usr/lib/security/methods.cfg
                            this option has to be used together with "--lib_name centrifyda"
    --not_joined            indicates the machine is not joined
                            this option has to be used together with "--lib_name centrifyda"
    --do_revert             set it if want to revert our changes
END
    exit 1;
}

use constant DEFAULT_CFG_FILE => "/usr/lib/security/methods.cfg";
use constant DEFAULT_CDC_TAG => "centrifydc";

my $cfgName = DEFAULT_CFG_FILE;
my $cfgHandler;
my $newName;
my $newHandler;

my $lib_name = DEFAULT_CDC_TAG;
my $is_method_hooked;
my $is_method_normal;
my $not_joined = 0;
my $do_revert;

my $cdc_stanza = << "END";
CENTRIFYDC:
        program = /usr/lib/security/CENTRIFYDC
        program_64 = /usr/lib/security/CENTRIFYDC64
        options = noprompt

END

my $cdc_commented_stanza = << "END";
*CENTRIFYDAMODIFY CENTRIFYDC:
*CENTRIFYDAMODIFY       program = /usr/lib/security/CENTRIFYDC
*CENTRIFYDAMODIFY       program_64 = /usr/lib/security/CENTRIFYDC64
*CENTRIFYDAMODIFY       options = noprompt

END

my $cda_stanza = << "END";
CENTRIFYDC:
        program = /usr/lib/security/CENTRIFYDA
        program_64 = /usr/lib/security/CENTRIFYDA64
        options = noprompt

END

my $cdalocal_stanza = << "END";
CENTRIFYDALOCAL:
        program = /usr/lib/security/CENTRIFYDALOCAL
        program_64 = /usr/lib/security/CENTRIFYDALOCAL64
        options = noprompt

END

#
#
# Check if "CENTRIFYDALOCAL" is found in methods.cfg
#
# $_[0]: the array ref contains all lines of methods.cfg
#
# Return value:
#       0 - Means that it is found
#       1 - Means that it is not found
#
sub isLamHookedMethod
{
    my ($cfgLines) = @_;
    
    foreach my $line (@$cfgLines)
    {
        if ($line =~ /^CENTRIFYDALOCAL/)
        {
            return 0;
        }
    }
    
    1;
}

#
#
# Check if methods config is as our expectation in methods.cfg
#
# $_[0]: the array ref contains all lines of methods.cfg
#
# Return value:
#       0 - Means normal
#       1 - Means abnormal
#
sub isLamNormalMethod
{
    my ($cfgLines) = @_;
    my $expectLines = "";
    
    if ($not_joined)
    {
        # CDC method is not enabled
        $expectLines .= $cdalocal_stanza;
    }
    else
    {
        # CDC method is enabled
        $expectLines .= $cdc_commented_stanza;
        $expectLines .= $cda_stanza;
        $expectLines .= $cdalocal_stanza;  
    }
    
    my @expectLines = split "\n", $expectLines;
    
    foreach my $expectLine (@expectLines)
    {
        if($expectLine eq "")
        {
            next;
        }
        
        $expectLine =~ s/\*/\\*/g;
        $expectLine =~ s/\s+/\\s*/g;
        my $found = 0;
        
        foreach my $line (@$cfgLines)
        {
            if ($line =~ /^\s*$expectLine\s*$/)
            {
                $found = 1;
                last;
            }
        }
        
        if(!$found)
        {
            return 1;
        }
    }
    
    0;
}

GetOptions( "lib_name=s" => \$lib_name,
            "cfg_file=s" => \$cfgName,
            "is_method_hooked" => \$is_method_hooked,
            "is_method_normal" => \$is_method_normal,
            "not_joined" => \$not_joined,
            "do_revert" => \$do_revert,
            ) or usage;

my $flock = autoedit_file_lock::new("$cfgName");
if (!defined $flock)
{
    ERROR_OUT("cannot lock file $cfgName");
    exit 1;
}

$cfgName = TraverseSymLink($cfgName);
if (!defined $cfgName)
{
    ERROR_OUT("meet error when check symlink $cfgName");
    exit 1;
}

my $cfgLines = ReadFile("$cfgName");
if (!defined $cfgLines)
{
    ERROR_OUT("cannot read $cfgName");
    exit 1;
}

my $cfgfile_empty = IsFileEmpty($cfgName);
if ($cfgfile_empty)
{
    exit 2;
}

# make sure the file ends with a blank line
if ($cfgLines !~ /\n\n$/)
{
    $cfgLines .= "\n";
}

my @cfgLines = split "\n", $cfgLines;

if ($is_method_hooked)
{
    if ($lib_name eq "centrifyda")
    {
        my $ret = isLamHookedMethod(\@cfgLines);
        
        exit $ret;
    }
    else
    {
        usage;
    }
}

if ($is_method_normal)
{
    if ($lib_name eq "centrifyda")
    {
        my $ret = isLamNormalMethod(\@cfgLines);
        
        exit $ret;
    }
    else
    {
        usage;
    }
}

if (!$do_revert)
{
    if ($cfgLines =~ /program.*\b\U$lib_name\E\b/)
    {
        # $lib_name is running
        # nothing else to do
        DEBUG_OUT("$lib_name already enabled, no-op");
        exit 0;
    }

    if ($lib_name eq "centrifydc")
    {
        my $newLines = "";

        # check if CDA is enabled
        if ($cfgLines =~ /^\s*program.*\bCENTRIFYDALOCAL\b/m)
        {
            # add CDA stanza
            $cfgLines =~ s/^CENTRIFYDALOCAL:\n(^[ \t].*\n)*(^\n)?//m;
            $cfgLines =~ s/^CENTRIFYDC:\n(^[ \t].*\n)*(^\n)?//m;
            $cfgLines .= $cdc_commented_stanza;
            $cfgLines .= $cda_stanza;
            $cfgLines .= $cdalocal_stanza;
        }
        else
        {
            $cfgLines .= $cdc_stanza;
        }
    }
    elsif ($lib_name eq "centrifyda")
    {
        # check if CDC is enabled
        if ($cfgLines =~ /^\s*program.*\bCENTRIFYDC\b/m)
        {
            # comment CDC stanza
            # add CDC stanza for DA
            # add CDA stanza
            $cfgLines =~ s/^CENTRIFYDC:\n(^[ \t].*\n)*(^\n)?//m;
            $cfgLines .= $cdc_commented_stanza;
            $cfgLines .= $cda_stanza;
            $cfgLines .= $cdalocal_stanza;
        }
        else
        {
            # add CDA stanza if DA not already enabled
            if ($cfgLines !~ /program.*\b\CENTRIFYDALOCAL\b/)
            {
                $cfgLines .= $cdalocal_stanza;
            }
            else
            {
                # DA is already enabled
                # nothing else to do
                DEBUG_OUT("$lib_name already enabled, no-op");
                exit 0;
            }
        }
    }
}
else
{
    if ($lib_name eq "centrifydc")
    {
        if ($cfgLines !~ /program.*\b\CENTRIFYDC\b/)
        {
            # $lib_name is not running
            # nothing else to do
            # if conf file is empty, show error
            DEBUG_OUT("$lib_name already disabled, no-op");
            exit 0;
        }
        if ($cfgLines =~ /CENTRIFYDAMODIFY/)
        {
            # Remove CENTRIFYDAMODIFY if any
            # Remove CENTRIFYDC stanza
            $cfgLines =~ s/^.*CENTRIFYDAMODIFY.*\n\n?//mg;
            $cfgLines =~ s/^CENTRIFYDC:\n(^[ \t].*\n)*(^\n)?//m;
        }
        else
        {
            $cfgLines =~ s/^CENTRIFYDC:\n(^[ \t].*\n)*(^\n)?//m;
        }
    }
    elsif ($lib_name eq "centrifyda")
    {
        if ($cfgLines !~ /program.*\b\CENTRIFYDALOCAL\b/)
        {
            # $lib_name is not running
            # nothing else to do
            DEBUG_OUT("$lib_name already disabled, no-op");
            exit 0;
        }
        $cfgLines =~ s/^CENTRIFYDALOCAL:\n(^[ \t].*\n)*(^\n)?//m;
        $cfgLines =~ s/^CENTRIFYDC:\n(^[ \t].*\n)*(^\n)?//m;

        if ($cfgLines =~ /CENTRIFYDAMODIFY/)
        {
            $cfgLines =~ s/^.*CENTRIFYDAMODIFY ?//mg;
        }
    }
}

# TODO: need to tell /etc/security/user we have reverted our settings
# pre-process
# rename "$newName", "$cfgName.$$";
# if ($do_revert)
# {
    # tell /etc/security/user we have reverted our settings
#    my $info = `/usr/share/centrifydc/scripts/da/autoedit_default_user.pl -cfg_file /tmp/user`;
#    print "$info\n";
#    print "autoedit_default_user.pl return $?\n";
# }
# rename "$cfgName.$$", "$newName";

my $backup = "$cfgName.pre_cdc";
if ($lib_name eq "centrifyda")
{
    $backup = "$cfgName.pre_cda";
}

if (!$do_revert)
{
    # don't backup empty config file
    if (!$cfgfile_empty)
    {
        if (!CopyFile($cfgName, $backup))
        {
            ERROR_OUT("Cannot backup $cfgName");
            exit 1;
        }
    }
    else
    {
        ERROR_OUT("Cannot backup $cfgName because it is empty");
        exit 1;
    }
}
else
{
    # install.sh and rpm/pkg scripts may override config files with
    # the backup files if found them exist.
    # To avoid affecting these code logic, we'd better delete
    # backup file after "do_revert"
    if (-e $backup && !$cfgfile_empty)
    {
        unlink $backup;
    }
}

# Do not make it worse writing to empty config file
# (empty config file already caused "internal injuries")
if (!$cfgfile_empty)
{
    if (!WriteFile($cfgName, $cfgLines))
    {
        ERROR_OUT("Cannot write $cfgName");
        exit 1;
    }
}
else
{
    ERROR_OUT("Cannot write $cfgName because it is empty");
    exit 1;
}

exit 0;

