#!/bin/sh /usr/share/centrifydc/perl/run.cda

##############################################################################
#
# Copyright (C) 2014-2019 Centrify Corporation. All rights reserved.
#
# This script is used to modify /etc/security/user
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error (general error)
#       2 if target file is empty
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts/da";
use CentrifyDA::Include;
use Getopt::Long qw(GetOptions);

use autoedit_common_utils qw(loadSetting getValueList);
use autoedit_file_lock;
use CentrifyDA::General qw(:debug :system);

sub usage
{
    warn << "END";
    This script is used to modify /etc/security/user
    --cfg_file              the full path of nsswitch.conf
    --lib_name              the name of centrifydc or centrifyda lib
    --is_user_hooked        check if "CENTRIFYDALOCAL" and "CENTRIFYMODIFY" are found in /etc/security/user
                            this option has to be used together with "--lib_name centrifyda"
    --is_user_normal        check if LAM config is as our expectation in /etc/security/user when DA LAM is enabled.
                            this option has to be used together with "--lib_name centrifyda"
    --not_joined            indicates the machine is not joined
                            this option has to be used together with "--lib_name centrifyda"
    --do_cover_user_root    set it if want to enable our module
                            to control user root
    --do_load_cdc_conf      set it if want to load the above settings
                            from centrifydc.conf
    --do_revert             set it if want to revert our changes
END
    exit 1;
}

use constant DEFAULT_CFG_FILE => "/etc/security/user";
use constant DEFAULT_CDC_TAG => "centrifydc";

my $cfgName = DEFAULT_CFG_FILE;
my $cfgHandler;
my $newName;
my $newHandler;

my $lib_name = DEFAULT_CDC_TAG;
my $is_user_hooked;
my $is_user_normal;
my $not_joined = 0;
my $do_load_cdc_conf;
my $do_cover_user_root;
my $do_revert;

my $cdc_template = "CENTRIFYDC OR CENTRIFYDC[NOTFOUND] AND (SYS)";
my $cda_template = "(CENTRIFYDALOCAL AND (SYS)) OR (CENTRIFYDALOCAL[NOTFOUND] AND (SYS))";
my $cdc_cda_template = "CENTRIFYDC OR (CENTRIFYDC[NOTFOUND] AND (SYS))";

#
#
# Check if both "CENTRIFYDALOCAL" and "*CENTRIFYDAMODIFY" can be found in /etc/security/user
#
# $_[0]: the array ref contains all lines of /etc/security/user
#
# Return value:
#       0 - Means that both "CENTRIFYDALOCAL" and "*CENTRIFYDAMODIFY" are found
#       1 - means that "CENTRIFYDALOCAL" or "*CENTRIFYDAMODIFY" is not found
#
sub isLamHookedUser
{
    my ($cfgLines) = @_;
    my $commentUserFound = 0;
    my $localSystemFound = 0;
    
    foreach my $line (@$cfgLines)
    {
        if (!$commentUserFound && $line =~ /CENTRIFYDALOCAL/)
        {
            $commentUserFound = 1;
        }
        
        if (!$localSystemFound && $line =~ /\*CENTRIFYDAMODIFY/)
        {
            $localSystemFound = 1;
        }
        
        if ($commentUserFound && $localSystemFound)
        {
            return 0;
        }
    }

    1;
}

#
# Get O/S SYSTEM setting
# 
sub getOriginalSystemSetting
{
    my ($stanza) = @_;
    my $modified = 0;
    my $setting = "";

    if ($stanza =~ /^.*CENTRIFYDAMODIFY.*SYSTEM\s*=(.*)$/m)
    {
        $setting = $1;
        $modified = 1;
    }
    else
    {
        if ($stanza =~ /^.*SYSTEM\s*=(.*)$/m)
        {
            $setting = $1;
        }
    }
    if ($setting =~ /\bCENTRIFYDC\b/)
    {
        $setting =~ s/CENTRIFYDC\[\w*\]\s*(OR|AND)\s*//ig;
        $setting =~ s/\s*(OR|AND)\s*CENTRIFYDC\[\w*\]//ig;
        $setting =~ s/CENTRIFYDC\s*(OR|AND)\s*//ig;
        $setting =~ s/\s*(OR|AND)\s*CENTRIFYDC//ig;

        # remove brackets if have.
        # I want to remove the pair of brackets, 
        # ONLY if they are the pair at head and end
        my $tmp = $setting;
        $tmp =~ s/\s//g;
        while ($tmp =~ s/\([^()]*\)//g)
        {}
        if (!$tmp)
        {
            $setting =~ s/^\s*\(//;
            $setting =~ s/\)\s*$//;
        }
    }
    $setting =~ s/\"//g;
    $setting =~ s/\(//;
    $setting =~ s/\)//;
    $setting =~ s/^\s+//;

    return ($setting);
}

#
#
# Check if DA LAM config is as our expectation in /etc/security/user when DA LAM is enabled.
#
# $_[0]: the array ref contains all lines of /etc/security/user
#
# Return value:
#       0 - Means normal
#       1 - means abnormal
#
sub isLamNormalUser
{
    my ($cfgLines) = @_;   
    my $setting  = "";
    my $system = "";
    my $comment = "";
    my $systemFound = 0;
    my $commentFound = 0;
    
    #determine what the config we expect is.
    
    # only DA enabled.
    if ($not_joined)
    {
        $system = "$cda_template";
        $comment = " *CENTRIFYDAMODIFY  SYSTEM = \"SYS\" ";
    }
    # CDC and CDA all enabled
    else
    {
        $setting = "$cda_template";
        $system = "$cdc_cda_template";
        $system =~ s/SYS/$setting/g;
        $comment = " *CENTRIFYDAMODIFY SYSTEM = \"$cdc_template\" ";
    }
    
    $system =~ s/SYS/\\w+/g;
    $system =~ s/\(/\\(/g;
    $system =~ s/\)/\\)/g;
    $system =~ s/\[/\\[/g;
    $system =~ s/\]/\\]/g;
    $system =~ s/\s+/\\s*/g;
    
    $comment =~ s/SYS/\\w+/g;
    $comment =~ s/\(/\\(/g;
    $comment =~ s/\)/\\)/g;
    $comment =~ s/\[/\\[/g;
    $comment =~ s/\]/\\]/g;
    $comment =~ s/\*/\\*/g;
    $comment =~ s/\s+/\\s*/g;
    
    foreach my $line (@$cfgLines)
    {
        if (!$systemFound && $line =~ /^\s*SYSTEM\s*=\s*\"$system\"/)
        {
            $systemFound = 1;
        }
        if (!$commentFound && $line =~ /^$comment$/)
        {
            $commentFound = 1;
        }
        
        if($systemFound && $commentFound)
        {
            return 0;
        }
    }

    1;
}

GetOptions( "cfg_file=s" => \$cfgName,
            "lib_name=s" => \$lib_name,
            "is_user_hooked" => \$is_user_hooked,
            "is_user_normal" => \$is_user_normal,
            "not_joined" => \$not_joined,
            "do_load_cdc_conf" => \$do_load_cdc_conf,
            "do_cover_user_root" => \$do_cover_user_root,
            "do_revert" => \$do_revert,
            ) or usage;

if (!$do_revert)
{
    if ($do_load_cdc_conf && !defined $do_cover_user_root)
    {
        my $do_cover_user_root = getValueList("adclient.autoedit.user.root");
        if (defined $do_cover_user_root 
            && defined $do_cover_user_root->[0]
            && $do_cover_user_root->[0] =~ /\btrue\b/s)
        {
            $do_cover_user_root = 1;
        }
        else
        {
            $do_cover_user_root = 0;
        }
    }
}

my $flock = autoedit_file_lock::new("$cfgName");
if (!defined $flock)
{
    ERROR_OUT("cannot lock file $cfgName");
    exit 1;
}

$cfgName = TraverseSymLink($cfgName);
if (!defined $cfgName)
{
    ERROR_OUT("meet error when check symlink $cfgName");
    exit 1;
}

my $cfgLines = ReadFile("$cfgName");
if (!defined $cfgLines)
{
    ERROR_OUT("cannot read $cfgName");
    exit 1;
}

my $cfgfile_empty = IsFileEmpty($cfgName);
if ($cfgfile_empty)
{
    exit 2;
}

my @cfgLines = split "\n", $cfgLines;

if ($is_user_hooked)
{
    if ($lib_name eq "centrifyda")
    {
       my $ret = isLamHookedUser(\@cfgLines);
        
        exit $ret;
    }
    else
    {
        usage;
    }
}

if ($is_user_normal)
{
    if ($lib_name eq "centrifyda")
    {
       my $ret = isLamNormalUser(\@cfgLines);
        
       exit $ret;
    }
    else
    {
        usage;
    }
}

# deal with "default" stanza
# The valid stanza head is a user name, followed by a :,
# not allowed empty string before user name.
if ($cfgLines =~ /^default:[^:]*\n/m)
{
    my $cdc_enabled = 0;
    my $cda_enabled = 0;
    my $changed = 0;
    my $preface = $`;
    my $stanza = $&;
    my $suffix = "\n$'";
    $stanza =~ s/\n+$/\n/;

    my $sysos = "";
    my $setting  = "";
    my $system = "";

    # Extract O/S SYSTEM setting
    $sysos = getOriginalSystemSetting($stanza);
    if (!$sysos)
    {
        # no SYSTEM setting, add it
        $stanza .= "\tSYSTEM = \"compat\"\n";
    }

    # Determine if CDC/CDA is enabled
    if ($stanza =~ /\bCENTRIFYDC\b/)
    {
        $cdc_enabled = 1;
    }
    if ($stanza =~ /\bCENTRIFYDALOCAL/)
    {
        $cda_enabled = 1;
    }

    # "--not_joined" is now actually "--do_revert" for when adleave is
    # executed and user config needs to be cleaned up.
    #       case: CDC enabled, DA enabled
    #             remove CDC from SYSTEM
    #       no-op for all other cases
    if ($lib_name eq "centrifyda")
    {
        if ($not_joined)
        {
            if ($cdc_enabled && $cda_enabled)
            {
                $do_revert = 1;
            }
            else
            {
                # no-op, just exit
                DEBUG_OUT("not joined to domain, no-op");
                exit 0;
            }
        }
    }

    # Processing enable/disable directive...
    if ($do_revert)
    {
        # Disable CDC or DA.  The following scenarios can happen:
        # 1. disable CDC (DA already disabled): via adleave
        # 2. disable DA (CDC already enabled): via dacontrol -d (already joined)
        # 3. disable DA (CDC already disabled): via dacontrol -d (not joined)
        # 4. disable CDC (DA already enabled): via adleave
        if ($lib_name eq "centrifydc")
        {
            if ($cdc_enabled)
            {
                if ($cda_enabled)
                {
                    # case: CDC enabled, DA enabled
                    # remove CDC from SYSTEM and from CENTRIFYDAMODIFY SYSTEM
                    $stanza =~ s/(^.*CENTRIFYDAMODIFY.*SYSTEM\s*=).*$/$1 \"$sysos\"/m;
                    $stanza =~ s/^\s*SYSTEM\s*=.*\n//m;
                    $setting = "$cda_template";
                    $setting =~ s/SYS/$sysos/g;
                    $stanza .= "\tSYSTEM = \"$setting\"\n";
                }
                else
                {
                    # case: CDC enabled, DA disabled
                    # remove CDC from SYSTEM to restore O/S setting
                    $stanza =~ s/(^\s*SYSTEM\s*=).*$/$1 \"$sysos\"/m;
                }
            }
            else
            {
                #print "DEBUG: CDC already disbled, NO-OP\n";
                DEBUG_OUT("$lib_name already enabled, no-op");
                exit 0;
            }
        }
        elsif ($lib_name eq "centrifyda")
        {
            if ($cda_enabled)
            {
                # case: CDC enabled/CDC disabled, DA enabled
                # remove SYSTEM and restore CENTRIFYDAMODIFY SYSTEM
                $stanza =~ s/^\s*SYSTEM\s*=.*\n//m;
                $stanza =~ s/^.*CENTRIFYDAMODIFY//m;
            }
            else
            {
                #print "DEBUG: DA already disbled, NO-OP\n";
                DEBUG_OUT("$lib_name already disabled, no-op");
                exit 0;
            }
        }
    }
    else
    {
        # Enable CDC or DA.  The following scenarios can happen:
        # 1. enable CDC (DA already disabled): via adjoin
        # 2. enable DA (CDC already enabled): via dacontrol -e (already joined)
        # 3. enable DA (CDC already disabled): via dacontrol -e (not joined)
        # 4. enable CDC (DA already enabled): via adjoin
        if ($lib_name eq "centrifydc")
        {
            if (!$cdc_enabled)
            {
                if (!$cda_enabled)
                {
                    # case: CDC not enabled, DA not enabled
                    # add CDC to SYSTEM
                    $setting = "$cdc_template";
                    $setting =~ s/SYS/$sysos/g;
                    $stanza =~ s/(^\s*SYSTEM\s*=).*$/$1 \"$setting\"/m;
                }
                else
                {
                    # case: CDC not enabled, DA enabled
                    # add CDC+DA to SYSTEM
                    $setting = "$cdc_template";
                    $setting =~ s/SYS/$sysos/g;
                    $stanza =~ s/(^.*CENTRIFYDAMODIFY.*SYSTEM\s*=).*$/$1 \"$setting\"/m;
                    $setting = "$cda_template";
                    $setting =~ s/SYS/$sysos/g;
                    $system = "$cdc_cda_template";
                    $system =~ s/SYS/$setting/g;
                    $stanza =~ s/(^\s*SYSTEM\s*=).*$/$1 \"$system\"/m;
                }
            }
            else
            {
                # CDC already enabled, DO NOTHING
                #print "DEBUG: CDC already enabled, NO-OP\n";
                DEBUG_OUT("$lib_name already enabled, no-op");
                exit 0;
            }
        }
        elsif ($lib_name eq "centrifyda")
        {
            if (!$cda_enabled)
            {
                if (!$cdc_enabled)
                {
                    # case CDC not enabled, DA not enabled
                    # add DA to SYSTEM
                    $stanza =~ s/(^\s*SYSTEM\s*=).*$/*CENTRIFYDAMODIFY$1 "$sysos"/m;
                    $setting = "$cda_template";
                    $setting =~ s/SYS/$sysos/g;
                    $stanza .= "\tSYSTEM = \"$setting\"\n";
                }
                else
                {
                    # case: CDC enabled, DA not enabled
                    # save old SYSTEM and add CDC+DA to SYSTEM
                    if ($stanza =~ /^\s*SYSTEM\s*=(.*)$/m)
                    {
                        $setting = $1;
                        $stanza =~ s/(^\s*SYSTEM\s*=).*$/*CENTRIFYDAMODIFY$1 $setting/m;
                    }
                    $setting = "$cda_template";
                    $setting =~ s/SYS/$sysos/g;
                    $system = "$cdc_cda_template";
                    $system =~ s/SYS/$setting/g;
                    $stanza .= "\tSYSTEM = \"$system\"\n";
                }
            }
            else
            {
                # CDA already enabled, DO NOTHING
                #print "DEBUG: DA already enabled, NO-OP\n";
                DEBUG_OUT("$lib_name already enabled, no-op");
                exit 0;
            }
        }
    }

    # construct updated "default" stanza
    $cfgLines = $preface.$stanza.$suffix;
}

if ($lib_name eq "centrifydc")
{
    # deal with "root" stanza
    if ($cfgLines =~ /^root:[^:]*\n/m)
    {
        my $preface = $`;
        my $stanza = $&;
        my $suffix = "\n$'";
        $stanza =~ s/\n+$/\n/;

        if ($do_revert)
        {
            # un-remark this line
            $stanza =~ s/\*\s*CENTRIFYDC ?//g;
        }
        elsif ($stanza =~ /^\s*SYSTEM\s*=/m)
        {
            if ($do_cover_user_root)
            {
                # remark this line to let "root" use default setting
                $stanza =~ s/^\s*SYSTEM\s*=/\* CENTRIFYDC $&/m;
            }
        }
        else
        {
            if (!$do_cover_user_root)
            {
                $stanza .= "\tSYSTEM = \"compat\"\n";
            }
        }

        $cfgLines = $preface.$stanza.$suffix;
    }
    elsif (!$do_cover_user_root)
    {
        $cfgLines .= "\nroot:\n"
                    ."\tSYSTEM = \"compat\"\n";
    }
}

# TODO: make sure methods.cfg have had the wanted lib_name
# pre-process
# if (!$do_revert)
# {
    # make sure methods.cfg have had the wanted lib_name
# }

my $backup = "$cfgName.pre_cdc";
if ($lib_name eq "centrifyda")
{
    $backup = "$cfgName.pre_cda";
}

if (!$do_revert)
{
    # don't backup empty config file
    if (!$cfgfile_empty)
    {
        if (!CopyFile($cfgName, $backup))
        {
            ERROR_OUT("Cannot backup $cfgName");
            exit 1;
        }
    }
    else
    {
        ERROR_OUT("Cannot backup $cfgName because it is empty");
        exit 1;
    }
}
else
{
    # install.sh and rpm/pkg scripts may override config files with
    # the backup files if found them exist.
    # To avoid affecting these code logic, we'd better delete
    # backup file after "do_revert"
    if (-e $backup && !$cfgfile_empty)
    {
        unlink $backup;
    }
}

# Do not make it worse writing to empty config file
# (empty config file already caused "internal injuries")
if (!$cfgfile_empty)
{
    if (!WriteFile($cfgName, $cfgLines))
    {
        ERROR_OUT("Cannot write $cfgName");
        exit 1;
    }
}
else
{
    ERROR_OUT("Cannot write $cfgName because it is empty");
    exit 1;
}

exit 0;

